`include "mips.h"

module alu(RSOperand, RTOperand, Immediate, UseImmediate, ALUOp, ALUout);
input [31:0] RSOperand;
input [31:0] RTOperand;
input [31:0] Immediate;
input UseImmediate;
input [7:0]	ALUOp;			
output reg [31:0] ALUout;
reg	[31:0] Temp;
reg [4:0] ShiftAmount;
wire [31:0] signed_div,signed_div_t,unsigned_div,signed_rem,signed_rem_t,unsigned_rem,signed_RS,signed_RT;
wire [32:0] unsigned_RS,unsigned_RT;
wire [63:0] signed_mult,signed_mult_t;
wire [63:0] unsigned_mult;
wire signofRS,signofRT;

assign signofRS = RSOperand[31];
assign signofRT = RTOperand[31];



   assign 	 unsigned_RS = {1'b0,RSOperand};
   assign 	 unsigned_RT = {1'b0,RTOperand};
   assign 	 signed_RS = (signofRS)?(~RSOperand)+1:RSOperand;
   assign 	 signed_RT = (signofRT)?(~RTOperand)+1:RTOperand;

assign signed_mult_t = signed_RS * signed_RT;
assign unsigned_mult = unsigned_RS * unsigned_RT;
assign unsigned_div = {1'b0,RSOperand}/{1'b0,RTOperand};
assign signed_div_t = {signed_RS}/{signed_RT};
assign unsigned_rem = {1'b0,RSOperand}%{1'b0,RTOperand};
assign signed_rem_t = {signed_RS}%{signed_RT};
assign signed_mult = (signofRS!=signofRT)?(~signed_mult_t) + 1:signed_mult_t;
assign signed_div = (signofRS!=signofRT)?(~signed_div_t) + 1:signed_div_t;
assign signed_rem = (signofRS)?(~signed_rem_t) + 1:signed_rem_t;

always @(*) begin
   Temp = (UseImmediate) ? Immediate : RTOperand;
   ShiftAmount = (UseImmediate) ? Immediate[10:6] : RSOperand[4:0];

   case (ALUOp)
      `select_alu_add:  ALUout = RSOperand + Temp;
      `select_alu_and:	ALUout = RSOperand & Temp;
      `select_alu_xor:	ALUout = RSOperand ^ Temp;
      `select_alu_or:   ALUout = RSOperand | Temp;
      `select_alu_nor:	ALUout = ~(RSOperand | Temp);
      `select_alu_sub:	ALUout = RSOperand - Temp;
      `select_alu_sltu:	ALUout = ({1'b0,RSOperand} < {1'b0,Temp}) ? 1 : 0;
      `select_alu_slt:	ALUout = (RSOperand[31] != Temp[31]) ? RSOperand[31] : ((RSOperand < Temp) ? 1 : 0);
      `select_alu_sra:  ALUout = {{32{RTOperand[31]}},RTOperand} >> ShiftAmount;
      `select_alu_srl:  ALUout = RTOperand >> ShiftAmount;
      `select_alu_sll:  ALUout = RTOperand << ShiftAmount;
	  `select_alu_mult_h: ALUout = signed_mult[63:32];
	  `select_alu_mult_l: ALUout = signed_mult[31:0];
	  `select_alu_multu_h: ALUout = unsigned_mult[63:32];
	  `select_alu_multu_l: ALUout = unsigned_mult[31:0];
	  `select_alu_div_h: ALUout = signed_rem;
      `select_alu_div_l: ALUout = signed_div;
	  `select_alu_divu_h: ALUout = unsigned_rem;
	  `select_alu_divu_l: ALUout = unsigned_div;
	  `select_alu_mfhi: ALUout = RSOperand;
	  `select_alu_mflo: ALUout = RSOperand;
     default:		ALUout = 0;
   endcase

  end

endmodule