`include "mips.h"

module branch_predictor(RESET,CLK,PC,Prediction,UpdateEnable,UpdatePC,UpdateValue,PredictorSelect,Direction);
input RESET;
input CLK;
input [31:0] PC;
output Prediction;
input UpdateEnable;
input [31:0] UpdatePC;
input UpdateValue;
input [2:0] PredictorSelect;
input Direction;

wire DPred;
wire GPred;
wire PPred;
wire TPred;
wire VPred;

reg [7:0] GHR;
reg [7:0] LHR [255:0];
reg [1:0] GSatCnt [255:0];
reg [1:0] PSatCnt [255:0];
reg [1:0] TSatCnt [255:0];

integer i;

assign DPred = (Direction)? 1'b1 : 1'b0;//If direction is backwards, predict taken, if direction is forward, predict not taken
assign GPred = GSatCnt[GHR][1];//Indexed by GHR, top most bit of sat cnt is prediction
assign PPred = PSatCnt[PC[9:2]][1];//Indexed by lower bits of PC (not the two LSB, =0),top most bit of sat cnt is prediction
assign TPred = TSatCnt[LHR[PC[9:2]]][1];//Index first table of history registers with pc, index sat cnt table with that history
assign VPred = ((TPred&GPred)|(TPred&PPred)|(GPred&PPred));

//Select the Prediction to Use
assign  Prediction = (PredictorSelect == `select_pred_dpred)?DPred:1'bz,
		Prediction = (PredictorSelect == `select_pred_gpred)?GPred:1'bz,
		Prediction = (PredictorSelect == `select_pred_ppred)?PPred:1'bz,
        Prediction = (PredictorSelect == `select_pred_tpred)?TPred:1'bz,
		Prediction = (PredictorSelect == `select_pred_vpred)?VPred:1'bz;

always @(posedge CLK or RESET)begin
	if(RESET)begin
		//Clear table on Reset;Shouldn't need to but simulator will complain undefined otherwise
		GHR = 0;
		for(i=0;i<256;i=i+1) LHR[i] = 0;
		for(i=0;i<256;i=i+1) GSatCnt[i] = 2'b01;
		for(i=0;i<256;i=i+1) PSatCnt[i] = 2'b01;
		for(i=0;i<256;i=i+1) TSatCnt[i] = 2'b01;
	end
	else if(UpdateEnable)begin
		//GPred
		case(GSatCnt[GHR])
		2'b00:begin
				if(UpdateValue)GSatCnt[GHR] = 2'b01;
				else GSatCnt[GHR] = 2'b00;
			  end
		2'b01:begin
				if(UpdateValue)GSatCnt[GHR] = 2'b10;
				else GSatCnt[GHR] = 2'b00;
			  end
		2'b10:begin
				if(UpdateValue)GSatCnt[GHR] = 2'b11;
				else GSatCnt[GHR] = 2'b01;
			  end
		2'b11:begin
				if(UpdateValue)GSatCnt[GHR] = 2'b11;
				else GSatCnt[GHR] = 2'b10;
			  end
		endcase
		GHR = (GHR << 1)|UpdateValue;
		
		//PPred
		case(PSatCnt[UpdatePC[9:2]])
		2'b00:begin
				if(UpdateValue)PSatCnt[UpdatePC[9:2]] = 2'b01;
				else PSatCnt[UpdatePC[9:2]] = 2'b00;
			  end
		2'b01:begin
				if(UpdateValue)PSatCnt[UpdatePC[9:2]] = 2'b10;
				else PSatCnt[UpdatePC[9:2]] = 2'b00;
			  end
		2'b10:begin
				if(UpdateValue)PSatCnt[UpdatePC[9:2]] = 2'b11;
				else PSatCnt[UpdatePC[9:2]] = 2'b01;
			  end
		2'b11:begin
				if(UpdateValue)PSatCnt[UpdatePC[9:2]] = 2'b11;
				else PSatCnt[UpdatePC[9:2]] = 2'b10;
			  end
		endcase		
		
		//TPred
		case(TSatCnt[LHR[UpdatePC[9:2]]])
		2'b00:begin
				if(UpdateValue)TSatCnt[LHR[UpdatePC[9:2]]] = 2'b01;
				else TSatCnt[LHR[UpdatePC[9:2]]] = 2'b00;
			  end
		2'b01:begin
				if(UpdateValue)TSatCnt[LHR[UpdatePC[9:2]]] = 2'b10;
				else TSatCnt[LHR[UpdatePC[9:2]]] = 2'b00;
			  end
		2'b10:begin
				if(UpdateValue)TSatCnt[LHR[UpdatePC[9:2]]] = 2'b11;
				else TSatCnt[LHR[UpdatePC[9:2]]] = 2'b01;
			  end
		2'b11:begin
				if(UpdateValue)TSatCnt[LHR[UpdatePC[9:2]]] = 2'b11;
				else TSatCnt[LHR[UpdatePC[9:2]]] = 2'b10;
			  end
		endcase	
		LHR[UpdatePC[9:2]] = (LHR[UpdatePC[9:2]]<<1)|UpdateValue;
		
	end
end



endmodule