`include "mips.h"

module issue_queue(RESET,CLK,fullFlush,maxIssue,queueNum,readyFlags,
					inOperation0,inImmediate0,inRT0,inRS0,inRD0,inQueue0,inROB0,inUImm0,
					inOperation1,inImmediate1,inRT1,inRS1,inRD1,inQueue1,inROB1,inUImm1,
					inOperation2,inImmediate2,inRT2,inRS2,inRD2,inQueue2,inROB2,inUImm2,
					inOperation3,inImmediate3,inRT3,inRS3,inRD3,inQueue3,inROB3,inUImm3,
					outOperation0,outImmediate0,outRT0,outRS0,outRD0,outROB0,outUImm0,
					outOperation1,outImmediate1,outRT1,outRS1,outRD1,outROB1,outUImm1,
					outReadyFlag0,outReadyFlag1,QueueTooFull,numIssued);

input RESET,CLK,fullFlush;
input [1:0] maxIssue,queueNum;
input [63:0] readyFlags;
output reg [63:0] outReadyFlag0;
output reg [63:0] outReadyFlag1;
input [7:0] inOperation0,inOperation1,inOperation2,inOperation3;
input [31:0] inImmediate0,inImmediate1,inImmediate2,inImmediate3;
input [4:0] inROB0,inROB1,inROB2,inROB3;
input [5:0] inRT0,inRT1,inRT2,inRT3,inRS0,inRS1,inRS2,inRS3,inRD0,inRD1,inRD2,inRD3;
input [1:0] inQueue0,inQueue1,inQueue2,inQueue3;
input inUImm0,inUImm1,inUImm2,inUImm3;
output reg outUImm0,outUImm1;
output reg[7:0] outOperation0,outOperation1;
output reg[31:0] outImmediate0,outImmediate1;
output reg[5:0] outRT0,outRT1,outRS0,outRS1,outRD0,outRD1;
output reg[4:0] outROB0,outROB1;
output reg QueueTooFull;

reg [31:0] max_fullness;
reg [31:0] total_occupancy;

reg [7:0] Operation [31:0];
reg [31:0] Immediate [31:0];
reg Valid [31:0];
reg [5:0] RT [31:0];
reg [5:0] RS [31:0];
reg [5:0] RD [31:0];
reg [4:0] ROB [31:0];
reg usesImm[31:0];
output reg [1:0] numIssued;
reg Inserted;
reg [5:0] FreeSlots;
//reg [31:0] wtfRS,wtfRT,wtfAND;
wire wtfRS,wtfRT,wtfAND;
integer i;

reg [31:0] entryReady;
always @(*)begin
for(i=0;i<32;i=i+1)entryReady[i] = Valid[i] & ((readyFlags>>RS[i])&1'b1) & ((readyFlags>>RT[i])&1'b1);
end

assign wtfRS = ((readyFlags>>RS[0])&1'b1);
assign wtfRT = ((readyFlags>>RT[0])&1'b1);
assign wtfAND = wtfRS & wtfRT;

always @(posedge CLK)begin
		//compact issue queue
	for(i=0;i<31;i=i+1)begin
		if((Valid[i]==0) && (Valid[i+1]==1))begin
			//$display("compacting %d to %d",i+1,i);
			Valid[i]=1;
			Operation[i] = Operation[i+1];
			Immediate[i] = Immediate[i+1];
			RT[i] = RT[i+1];
			RS[i] = RS[i+1];
			RD[i] = RD[i+1];
			ROB[i] = ROB[i+1];
			usesImm[i] = usesImm[i+1];
			Valid[i+1]=0;
		end
	end
end

always @(negedge CLK)begin
	if(RESET || fullFlush)begin
		for(i=0;i<32;i=i+1)Valid[i]=0;
		FreeSlots = 32;
		numIssued = 0;
		outReadyFlag0 = 0;
		outReadyFlag1 = 0;
		outOperation0 = 0;
		outImmediate0 = 0;
		outRT0 = 0;
		outRS0 = 0;
		outRD0 = 0;
		outROB0 = 0;
		outUImm0 = 0;
		outReadyFlag0 = 0;
		outOperation1 = 0;
		outImmediate1 = 0;
		outRT1 = 0;
		outRS1 = 0;
		outRD1 = 0;
		outUImm1 = 0;
		outROB1 = 0;
		outReadyFlag1 = 0;
		total_occupancy = 0;
		max_fullness = 0;
	end
	else begin

	
		//insert instructions
		Inserted = 0;
		for(i=0;i<32;i=i+1)begin
			if((Valid[i]==0) && (Inserted==0) && (inOperation0!=0) && (inQueue0==queueNum))begin
				Valid[i]=1;
				Inserted=1;
				Operation[i] = inOperation0;
				Immediate[i] = inImmediate0;
				RT[i] = inRT0;
				RS[i] = inRS0;
				RD[i] = inRD0;
				ROB[i] = inROB0;
				usesImm[i] = inUImm0;
				FreeSlots = FreeSlots - 1;
			end
		end
		
		Inserted = 0;
		for(i=0;i<32;i=i+1)begin
			if((Valid[i]==0) && (Inserted==0) && (inOperation1!=0) && (inQueue1==queueNum))begin
				Valid[i]=1;
				Inserted=1;
				Operation[i] = inOperation1;
				Immediate[i] = inImmediate1;
				RT[i] = inRT1;
				RS[i] = inRS1;
				RD[i] = inRD1;
				ROB[i] = inROB1;
				usesImm[i] = inUImm1;
				FreeSlots = FreeSlots - 1;
			end
		end
		
		Inserted = 0;
		for(i=0;i<32;i=i+1)begin
			if((Valid[i]==0) && (Inserted==0) && (inOperation2!=0) && (inQueue2==queueNum))begin
				Valid[i]=1;
				Inserted=1;
				Operation[i] = inOperation2;
				Immediate[i] = inImmediate2;
				RT[i] = inRT2;
				RS[i] = inRS2;
				RD[i] = inRD2;
				ROB[i] = inROB2;
				usesImm[i] = inUImm2;
				FreeSlots = FreeSlots - 1;
			end
		end
		
		Inserted = 0;
		for(i=0;i<32;i=i+1)begin
			if((Valid[i]==0) && (Inserted==0) && (inOperation3!=0) && (inQueue3==queueNum))begin
				Valid[i]=1;
				Inserted=1;
				Operation[i] = inOperation3;
				Immediate[i] = inImmediate3;
				RT[i] = inRT3;
				RS[i] = inRS3;
				RD[i] = inRD3;
				ROB[i] = inROB3;
				usesImm[i] = inUImm3;
				FreeSlots = FreeSlots - 1;
			end
		end

		total_occupancy = total_occupancy + (32-FreeSlots);
		if((32-FreeSlots)>max_fullness)max_fullness = (32-FreeSlots);

		//find next ones to issue		
		numIssued = 0;
		outReadyFlag0 = 0;
		outReadyFlag1 = 0;
		outOperation0 = 0;
		outImmediate0 = 0;
		outRT0 = 0;
		outRS0 = 0;
		outRD0 = 0;
		outROB0 = 0;
		outUImm0 = 0;
		outReadyFlag0 = 0;
		outOperation1 = 0;
		outImmediate1 = 0;
		outRT1 = 0;
		outRS1 = 0;
		outRD1 = 0;
		outUImm1 = 0;
		outROB1 = 0;
		outReadyFlag1 = 0;
		for(i=0;i<32;i=i+1)begin
			if(entryReady[i] && numIssued < maxIssue)begin
				if(numIssued == 0)begin
					outOperation0 = Operation[i];
					outImmediate0 = Immediate[i];
					outRT0 = RT[i];
					outRS0 = RS[i];
					outRD0 = RD[i];
					outROB0 = ROB[i];
					outUImm0 = usesImm[i];
					outReadyFlag0 = (1'b1 << RD[i]);
					numIssued = 1;					
					FreeSlots = FreeSlots + 1;
					Valid[i] = 0;
				end
				else begin
					outOperation1 = Operation[i];
					outImmediate1 = Immediate[i];
					outRT1 = RT[i];
					outRS1 = RS[i];
					outRD1 = RD[i];
					outUImm1 = usesImm[i];
					outROB1 = ROB[i];
					outReadyFlag1 = (1'b1 << RD[i]);
					numIssued = 2;									
					FreeSlots = FreeSlots + 1;
					Valid[i] = 0;
				end		
			end
		end
		
	end
	QueueTooFull = (FreeSlots < 4);
end

endmodule