 
//-------------------------------------------------------------------------
//
//  Copyright (c) 1999 Cornell University
//  Computer Systems Laboratory
//  Cornell University, Ithaca, NY 14853
//  All Rights Reserved
//
//  Permission to use, copy, modify, and distribute this software
//  and its documentation for any purpose and without fee is hereby
//  granted, provided that the above copyright notice appear in all
//  copies. Cornell University makes no representations
//  about the suitability of this software for any purpose. It is
//  provided "as is" without express or implied warranty. Export of this
//  software outside of the United States of America may require an
//  export license.
//
//  $Id: mem.v,v 1.8 2000/10/14 19:21:49 heinrich Exp $
//
//-------------------------------------------------------------------------

`include "mips.h"
`include "cache.h"
   
module mem (CLK, memOperation, isMem, RESET, MAR, Valid, SMDR, IaddrA, IaddrB, IaddrC, IaddrD, Bus, Read, Write, Addr, cacheOut, IinA, IinB, IinC, IinD, Istall, Dstall, isLoad);
   
   input		CLK;
   input [7:0]  memOperation;
   input		isMem;
   input		RESET;
   input [31:0] MAR;
   input		Valid;
   input [31:0] SMDR;
   input [31:0] IaddrA;
   input [31:0] IaddrB;
   input [31:0] IaddrC;
   input [31:0] IaddrD;
   
   inout [31:0] Bus;
   reg [31:0] 	BusReg;
   
   output		Read;		
   reg 			Read;
   output		Write;	
   reg 			Write;
   
   output reg [31:0] Addr;
   output [31:0] cacheOut;
   output [31:0] 	IinA;
   output [31:0] 	IinB;
   output [31:0] 	IinC;
   output [31:0] 	IinD;
   reg [31:0] preCacheOut;
   reg [`I_TAG_WIDTH-1:0] iTagA;
   reg [`I_TAG_WIDTH-1:0] iTagB;
   reg [`I_TAG_WIDTH-1:0] iTagC;
   reg [`I_TAG_WIDTH-1:0] iTagD;
   reg [`D_TAG_WIDTH-1:0] dTag;
   reg [1:0] iStateA;
   reg [1:0] iStateB;
   reg [1:0] iStateC;
   reg [1:0] iStateD;
   reg [1:0] dState;
   reg rDstall;
   reg rIstallA,rIstallB,rIstallC,rIstallD;
   reg deferred_rIstallA,deferred_rIstallB,deferred_rIstallC,deferred_rIstallD;
   reg retry;
   output reg isLoad;
   reg isStore;
   reg [`D_WO_WIDTH-1:0] offset;
   reg writeDone;
   reg readDone;
   reg [1:0] Dsize;
   reg [31:0] iCacheOutA;
   reg [31:0] iCacheOutB;
   reg [31:0] iCacheOutC;
   reg [31:0] iCacheOutD;
   reg [4:0] dcount;
   reg [4:0] icount;
   reg [31:0] num_icache_accesses;
   reg hit; 
   wire [31:0] magic_number;
   assign magic_number = 32'h0043BFF4;
   wire google;
   assign google = (IaddrA[31:5]==magic_number[31:5])|(IaddrB[31:5]==magic_number[31:5])|(IaddrC[31:5]==magic_number[31:5])|(IaddrD[31:5]==magic_number[31:5]);
  
   output Dstall;
   output Istall;
   
   // These wires are for performing sub-word stores
   wire [4:0] sa;
   wire [31:0] mask;
   wire [31:0] value;

   // This is a good place to keep track of the cache stats
   always @(posedge CLK) begin
      // This information is for STATs only
      if (RESET) begin
		 CPU.numLoads =`TICK 32'b0;
		 CPU.numStores =`TICK 32'b0;
      end
      
      if (isLoad & ~Dstall) begin
	 	CPU.numLoads =`TICK CPU.numLoads + 1;
      end
      if (isStore & ~Dstall) begin
	 	CPU.numStores =`TICK CPU.numStores + 1;
      end
   end
   
   // This always block handles simple decoding of isLoad or isStore
   always @(*) begin
	  if ((isMem)&&(memOperation==`select_mem_lw || memOperation==`select_mem_lh || memOperation==`select_mem_lhu || memOperation==`select_mem_lbu || memOperation==`select_mem_lb))begin
	 	isLoad = 1;
      end
      else begin
	 	isLoad = 0;
      end
      
	  if ((isMem)&&(memOperation==`select_mem_sw || memOperation==`select_mem_sh || memOperation==`select_mem_sb))begin
	 	isStore = 1;
      end
      else begin
      	isStore = 0;
      end
      
      // Set the data size correctly
      if (memOperation==`select_mem_sh) begin
		 Dsize = `SIZE_HALF;
      end
      else if (memOperation==`select_mem_sb) begin
		 Dsize = `SIZE_BYTE;
      end
      else begin
		 Dsize = `SIZE_WORD;
      end
   end
   
   // Main I$ control signals
    assign Istall = rIstallA|rIstallB|rIstallC|rIstallD|deferred_rIstallA|deferred_rIstallB|deferred_rIstallC|deferred_rIstallD;
    assign IinA = iCacheOutA;
    assign IinB = iCacheOutB;
	assign IinC = iCacheOutC;
	assign IinD = iCacheOutD;
   
   // Main D$ control signals
   assign Dstall = rDstall;

   wire [31:0] preLoadDataLB, preLoadDataLH;
   
   assign      preLoadDataLB = (preCacheOut >> (( ~MAR & 32'h3) << 3)) & 32'hff;
   assign      preLoadDataLH = (preCacheOut >> (((~MAR >> 1 ) & 32'h1)  << 4)) & 32'hffff;
   assign      
	       cacheOut = (memOperation == `select_mem_lw)  ?  preCacheOut : 32'bz,
	       cacheOut = (memOperation == `select_mem_lhu) ? (preCacheOut >> ((( ~MAR >> 1 ) & 32'h1) << 4)) & 32'hffff : 32'bz,
	       cacheOut = (memOperation == `select_mem_lbu) ? (preCacheOut >> ((~MAR & 32'h3) << 3)) & 32'hff : 32'bz,
	       cacheOut = (memOperation == `select_mem_lb)  ? {{24{preLoadDataLB[7]}}, preLoadDataLB[7:0]} : 32'bz,
	       cacheOut = (memOperation == `select_mem_lh)  ? {{16{preLoadDataLH[15]}},preLoadDataLH[15:0]}:32'bz;
  

   
   // This always block handles the D$ cache access, we trigger on I3 because unknown MAR's seemed to be causing problems
   // With the isStore||isLoad block we won't do more than triggering on MAR other than the if statement
   always @(*) begin
      if(isStore || isLoad)begin
        // This should set cacheOut appropriately
        // Read tag and state
      	dTag = $dcache_tag_read(MAR[`D_INDEX],0);
      	dState = $dcache_state_read(MAR[`D_INDEX],0);

		//DCache Hit, tags match, and is valid
      	if((dState[`D_VALID]==1'b1) && (dTag == MAR[`D_TAG]))begin
			hit =1'b1;
			if(isLoad)begin
				preCacheOut =$dcache_data_read(MAR[`D_INDEX],0,MAR[`D_WO]);
			end
	    end
	      //DCache Miss
	    else begin
			hit =1'b0;
			rDstall =1'b1;

			//Dirty and Valid, writeback to memory then read
			if(dState == 2'b11)begin
			  dcount =`TICK 4'b0000;
			end
			else begin
			//Not dirty so just read block from memory
			  dcount =`TICK 4'b1000;
			end
	    end
	  end
   end

   //Retry Logic - Will always be a hit
   //Handle I$ miss before D$ miss if both missed
   //Separated from other triggered cache always blocks for easier debugging/reading
   always @(posedge retry)begin
		if(rIstallA||rIstallB||rIstallC||rIstallD)begin
			iTagA = $icache_tag_read(IaddrA[`I_INDEX],0);
			iStateA = $icache_state_read(IaddrA[`I_INDEX],0);
			iTagB = $icache_tag_read(IaddrB[`I_INDEX],0);
			iStateB = $icache_state_read(IaddrB[`I_INDEX],0);
			iTagC = $icache_tag_read(IaddrC[`I_INDEX],0);
			iStateC = $icache_state_read(IaddrC[`I_INDEX],0);
			iTagD = $icache_tag_read(IaddrD[`I_INDEX],0);
			iStateD = $icache_state_read(IaddrD[`I_INDEX],0);
			//ICache Hit
			if((iStateA[`I_VALID]==1'b1) && (iTagA == IaddrA[`I_TAG])) begin
				iCacheOutA = $icache_data_read(IaddrA[`I_INDEX],0,IaddrA[`I_WO]);
			end
			else $finish;
		  	if((iStateB[`I_VALID]==1'b1) && (iTagB == IaddrB[`I_TAG])) begin
				iCacheOutB = $icache_data_read(IaddrB[`I_INDEX],0,IaddrB[`I_WO]);
			end
			else $finish;
			if((iStateC[`I_VALID]==1'b1) && (iTagC == IaddrC[`I_TAG])) begin
				iCacheOutC = $icache_data_read(IaddrC[`I_INDEX],0,IaddrC[`I_WO]);
			end
			else $finish;
			if((iStateD[`I_VALID]==1'b1) && (iTagD == IaddrD[`I_TAG])) begin
				iCacheOutD = $icache_data_read(IaddrD[`I_INDEX],0,IaddrD[`I_WO]);
			end
			else $finish;
		end
		else if(rDstall)begin
	      		dTag = $dcache_tag_read(MAR[`D_INDEX],0);
	      		dState = $dcache_state_read(MAR[`D_INDEX],0);
	      		//DCache Hit
				if((dState[`D_VALID]==1'b1) && (dTag == MAR[`D_TAG]))begin
					hit =1'b1;
					if(isLoad)begin
						preCacheOut =$dcache_data_read(MAR[`D_INDEX],0,MAR[`D_WO]);
					end
				end
				if(deferred_rIstallA)begin
					rIstallA = deferred_rIstallA;
					deferred_rIstallA = 0;
				end
				if(deferred_rIstallB)begin
					rIstallB = deferred_rIstallB;
					deferred_rIstallB = 0;
				end
				if(deferred_rIstallC)begin
					rIstallC = deferred_rIstallC;
					deferred_rIstallC = 0;
				end
				if(deferred_rIstallD)begin
					rIstallD = deferred_rIstallD;
					deferred_rIstallD = 0;
				end

		end
		retry =`TICK 0;
   end
  
   // This always block handles the I$ access
   always @(IaddrA or IaddrB or IaddrC or IaddrD) begin
        // This should set iCacheOut appropriately
		iTagA = $icache_tag_read(IaddrA[`I_INDEX],0);
		iStateA = $icache_state_read(IaddrA[`I_INDEX],0);
		iTagB = $icache_tag_read(IaddrB[`I_INDEX],0);
		iStateB = $icache_state_read(IaddrB[`I_INDEX],0);
		iTagC = $icache_tag_read(IaddrC[`I_INDEX],0);
		iStateC = $icache_state_read(IaddrC[`I_INDEX],0);
		iTagD = $icache_tag_read(IaddrD[`I_INDEX],0);
		iStateD = $icache_state_read(IaddrD[`I_INDEX],0);
		num_icache_accesses = num_icache_accesses + 1;
		//ICache Hit
		if((iStateA[`I_VALID]==1'b1) && (iTagA == IaddrA[`I_TAG])
		 &&(iStateB[`I_VALID]==1'b1) && (iTagB == IaddrB[`I_TAG])
		 &&(iStateC[`I_VALID]==1'b1) && (iTagC == IaddrC[`I_TAG])
		 &&(iStateD[`I_VALID]==1'b1) && (iTagD == IaddrD[`I_TAG])) begin
			iCacheOutA = $icache_data_read(IaddrA[`I_INDEX],0,IaddrA[`I_WO]);
			iCacheOutB = $icache_data_read(IaddrB[`I_INDEX],0,IaddrB[`I_WO]);
			iCacheOutC = $icache_data_read(IaddrC[`I_INDEX],0,IaddrC[`I_WO]);
			iCacheOutD = $icache_data_read(IaddrD[`I_INDEX],0,IaddrD[`I_WO]);
		end
		//ICache Miss, we don't have dirty here because you only read I$
		else begin
			iCacheOutA =32'b0;
			iCacheOutB =32'b0;
			iCacheOutC =32'b0;
			iCacheOutD =32'b0;
			if((iStateA[`I_VALID]!=1'b1) || (iTagA != IaddrA[`I_TAG]))begin
				if(rDstall)deferred_rIstallA = 1;
				else rIstallA =1'b1;
			end
			if(((iStateB[`I_VALID]!=1'b1) || (iTagB != IaddrB[`I_TAG]))
				&&(IaddrA[`I_INDEX]!=IaddrB[`I_INDEX])) begin
				if(rDstall)deferred_rIstallB = 1;
				else rIstallB =1'b1;
			end	
			if(((iStateC[`I_VALID]!=1'b1) || (iTagC != IaddrC[`I_TAG]))
				&&(IaddrA[`I_INDEX]!=IaddrC[`I_INDEX])
				&&(IaddrB[`I_INDEX]!=IaddrC[`I_INDEX])) begin 
				if(rDstall)deferred_rIstallC = 1;
				else rIstallC =1'b1;
			end		
			if(((iStateD[`I_VALID]!=1'b1) || (iTagD != IaddrD[`I_TAG]))
				&&(IaddrA[`I_INDEX]!=IaddrD[`I_INDEX])
				&&(IaddrB[`I_INDEX]!=IaddrD[`I_INDEX])
				&&(IaddrC[`I_INDEX]!=IaddrD[`I_INDEX])) begin
				if(rDstall)deferred_rIstallD = 1;
				else rIstallD =1'b1;
			end
			icount =0;
		end
   end

   //Handle All Miss State Machines Here, POSEDGE triggered
   always @(posedge CLK) begin
	//I$ miss
	if(rIstallA)begin
		//Set read and addr
		if(icount==0)begin	
			Read =1'b1;
			Addr =IaddrA;
		end
		//Wait for valid and read in words for 8 cycles
		if(Valid && (icount < 8))begin
			$icache_tag_write(IaddrA[`I_INDEX],0,IaddrA[`I_TAG]);
			$icache_state_write(IaddrA[`I_INDEX],0,2'b01);
			$icache_data_write(IaddrA[`I_INDEX],0,icount[2:0],Bus);
		end
		//After 8 Cycles retry cache read
		if(icount==8)begin
			Read =1'b0;
			if(rIstallB==0 && rIstallC==0 && rIstallD==0)retry =`TICK 1'b1;
			icount =`TICK icount+1;
			rIstallA =1'b0;
			CPU.numIMisses =CPU.numIMisses + 1;
			icount =`TICK 0;
		end
		//After 9 cycles we unstall and go on
		else if(icount == 9) begin

		end
		//Advance counter on valid signal
		else begin
			if(Valid)begin
				icount =`TICK icount+1;
			end
		end
	end
		//I$ miss
	else if(rIstallB)begin
		//Set read and addr
		if(icount==0)begin	
			Read =1'b1;
			Addr =IaddrB;
		end
		//Wait for valid and read in words for 8 cycles
		if(Valid && (icount < 8))begin
			$icache_tag_write(IaddrB[`I_INDEX],0,IaddrB[`I_TAG]);
			$icache_state_write(IaddrB[`I_INDEX],0,2'b01);
			$icache_data_write(IaddrB[`I_INDEX],0,icount[2:0],Bus);
		end
		//After 8 Cycles retry cache read
		if(icount==8)begin
			Read =1'b0;
			if(rIstallC==0 && rIstallD==0)retry =`TICK 1'b1;
			icount =`TICK icount+1;
			rIstallB =1'b0;
			CPU.numIMisses =CPU.numIMisses + 1;
			icount =`TICK 0;
		end
		//After 9 cycles we unstall and go on
		else if(icount == 9) begin

		end
		//Advance counter on valid signal
		else begin
			if(Valid)begin
				icount =`TICK icount+1;
			end
		end
	end
	else if(rIstallC)begin
		//Set read and addr
		if(icount==0)begin	
			Read =1'b1;
			Addr =IaddrC;
		end
		//Wait for valid and read in words for 8 cycles
		if(Valid && (icount < 8))begin
			$icache_tag_write(IaddrC[`I_INDEX],0,IaddrC[`I_TAG]);
			$icache_state_write(IaddrC[`I_INDEX],0,2'b01);
			$icache_data_write(IaddrC[`I_INDEX],0,icount[2:0],Bus);
		end
		//After 8 Cycles retry cache read
		if(icount==8)begin
			Read =1'b0;
			if(rIstallD==0)retry =`TICK 1'b1;
			icount =`TICK icount+1;
			rIstallC =1'b0;
			CPU.numIMisses =CPU.numIMisses + 1;
			icount =`TICK 0;
		end
		//After 9 cycles we unstall and go on
		else if(icount == 9) begin

		end
		//Advance counter on valid signal
		else begin
			if(Valid)begin
				icount =`TICK icount+1;
			end
		end
	end
	else if(rIstallD)begin
		//Set read and addr
		if(icount==0)begin	
			Read =1'b1;
			Addr =IaddrD;
		end
		//Wait for valid and read in words for 8 cycles
		if(Valid && (icount < 8))begin
			$icache_tag_write(IaddrD[`I_INDEX],0,IaddrD[`I_TAG]);
			$icache_state_write(IaddrD[`I_INDEX],0,2'b01);
			$icache_data_write(IaddrD[`I_INDEX],0,icount[2:0],Bus);
		end
		//After 8 Cycles retry cache read
		if(icount==8)begin
			Read =1'b0;
			retry =`TICK 1'b1;
			icount =`TICK icount+1;
			rIstallD =1'b0;
			CPU.numIMisses =CPU.numIMisses + 1;
			icount =`TICK 0;
		end
		//After 9 cycles we unstall and go on
		else if(icount == 9) begin

		end
		//Advance counter on valid signal
		else begin
			if(Valid)begin
				icount =`TICK icount+1;
			end
		end
	end
	//D$ miss
	else if(rDstall) begin
	  //After writeback is done (if needed) read in the block, setting read and address
	  if(dcount==8)begin	
		Read =1'b1;
		Addr =MAR;
		Write =`TICK 1'b0;
	  end
	  //Writeback block from cache to memory
	  if(dcount<8) begin
		Write =1'b1;
		Addr ={dTag,MAR[`D_INDEX],dcount[2:0],2'b00};
		BusReg =$dcache_data_read(MAR[`D_INDEX],0,dcount[2:0]);
		dcount =`TICK dcount+1;
	  end
	  //Read block from memory to cache
	  else if(Valid && (dcount < 16)) begin
		$dcache_tag_write(MAR[`D_INDEX],0,MAR[`D_TAG]);
		$dcache_state_write(MAR[`D_INDEX],0,2'b01);
		$dcache_data_write(MAR[`D_INDEX],0,dcount[2:0],Bus);
	        dcount =`TICK dcount+1;	
	  end
	  //Reread data after its updated
	  else if(dcount == 16) begin
		Read =1'b0;
		retry =`TICK 1'b1;
		dcount =`TICK dcount+1;
		rDstall =`TICK 1'b0;
		CPU.numDMisses =CPU.numDMisses + 1;
		dcount =`TICK 0;
	  end
	  //Unstall and continue pipe
	  else if(dcount == 17) begin

          end
		
	end
   end
   
   //Shift Amount
   assign sa = (Dsize == `SIZE_BYTE) ? (( ~MAR & 32'h3) << 3) : 5'bz,
     	  sa = (Dsize == `SIZE_HALF) ? (((~MAR >> 1) & 32'h1) << 4) : 5'bz,
     	  sa = (Dsize == `SIZE_WORD) ? 0 : 5'bz;
   //Data Mask
   assign 
	  mask = (Dsize == `SIZE_BYTE) ? (32'hff << sa) : 32'bz,
	  mask = (Dsize == `SIZE_HALF) ? (32'hffff << sa) : 32'bz,
	  mask = (Dsize == `SIZE_WORD) ? 0 : 32'bz;
   
   // Assume that stores write the data array on the negedge of the clock
   always @(negedge CLK) begin   
      // This code is complete and it works!  Note how it uses the 
      // cache PLI calls for both reads and writes
      if (hit & isStore) begin
	 // Need to write the data array and the Dirty bit
	 // Handle sub-word stores here
	 if (Dsize == `SIZE_WORD) begin
	    $dcache_data_write(MAR[`D_INDEX], 0, MAR[`D_WO], SMDR);
	 end
	 else if (Dsize == `SIZE_HALF) begin
	    $dcache_data_write(MAR[`D_INDEX], 0, MAR[`D_WO], ($dcache_data_read(MAR[`D_INDEX],0,MAR[`D_WO]) & ~mask)|((SMDR & 32'hffff) << sa));
	 end
	 else if (Dsize == `SIZE_BYTE) begin
	    $dcache_data_write(MAR[`D_INDEX], 0, MAR[`D_WO], ($dcache_data_read(MAR[`D_INDEX],0,MAR[`D_WO]) & ~mask)|((SMDR & 32'hff) << sa));
	 end      
	 // Write Dirty and Valid bits
	 $dcache_state_write(MAR[`D_INDEX], 0, 2'b11);            
      end
   end
   
   // This always block handles reset
   // We moved miss state machines out of this block for easier reading/debugging
   always @(posedge CLK) begin
      if (RESET) begin
	 	offset =`TICK 0;
	 	retry =`TICK 1'b0;
	 	writeDone =`TICK 1'b0;
	 	readDone =`TICK 1'b0;
	 	CPU.numDMisses =`TICK 32'b0;
	 	CPU.numIMisses =`TICK 32'b0;
		Read = 1'b0;
		Write = 1'b0;
		//Addr = 32'b0;
		rDstall = 0;
		icount = 0;
		dcount = 0;
		rIstallA = 1;
		rIstallB = 0;
		rIstallC = 0;
		rIstallD = 0;
		deferred_rIstallA = 0;
		deferred_rIstallB = 0;
		deferred_rIstallC = 0;
		deferred_rIstallD = 0;
		isStore = 0;
		isLoad = 0;
		preCacheOut = 0;
		num_icache_accesses = 0;
      end
   end // always @ (posedge CLK)

   //Initialize some signals
   initial begin
	Read = 1'b0;
	Write = 1'b0;
	Addr = 32'b0;
	rDstall = 0;
	icount = 0;
	dcount = 0;
	rIstallA = 0;
	rIstallB = 0;
	rIstallC = 0;
	rIstallD = 0;
	isStore = 0;
	isLoad = 0;
   end
   
   // Only drive Bus during write back
   assign Bus = (Write) ? BusReg : 32'bz;
   
   
//--------------------------------------------------------------------  

endmodule			
